#!/usr/bin/env python3

import argparse
from collections.abc import Iterable

from . import model

CONNECTION_ANY = 'any'
CONNECTION_EXTERNAL = 'external'
ALL_CONNECTION_VALUES = (CONNECTION_ANY, CONNECTION_EXTERNAL) + model.ALL_CONNECTIONS
possible_connections = ','.join(f'{ac}' for ac in ALL_CONNECTION_VALUES)


def connection(connections: str) -> 'list[str]':
	group: 'Iterable[str]'
	out = []
	for c in connections.split(','):
		if c == CONNECTION_ANY:
			group = model.ALL_CONNECTIONS
		elif c == CONNECTION_EXTERNAL:
			group = model.EXTERNAL_CONNECTIONS
		elif c in model.ALL_CONNECTIONS:
			group = [c]
		else:
			raise argparse.ArgumentTypeError(f'invalid connection {c!r}, should be one of {possible_connections}')

		for ci in group:
			if ci not in out:
				out.append(ci)
	return out



#!/usr/bin/env python3

import subprocess
import re
import typing
import shlex

from . import model


def is_available() -> bool:
	cmd = ['which', 'xrandr']
	try:
		subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		return True
	except:
		return False


class Api(model.Api):

	encoding = 'utf-8'


	# ---------- meta ----------

	def print_version(self) -> None:
		subprocess.call(['xrandr', '--version'])


	# ---------- read ----------

	re_geometry = r'(?P<width>[0-9]+)x(?P<height>[0-9]+)\+(?P<offsetx>[0-9]+)\+(?P<offsety>[0-9]+)'
	re_connected = r'^(?P<name>[^\s]+) connected (primary )?(%s)?' % re_geometry
	reo_connected = re.compile(re_connected)
	re_available_resolution = r'\s+(?P<width>[0-9]+)x(?P<height>[0-9]+)\s+([0-9]+\.[0-9]+)((?P<on>\*)| )((?P<default>\+)| )'
	reo_available_resolution = re.compile(re_available_resolution)

	def _get_xrandr_output(self) -> typing.Sequence[str]:
		cmd = ['xrandr']
		p = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
		return p.stdout.decode(self.encoding).splitlines()

	def _read_test_file(self, fn: str) -> typing.Sequence[str]:
		with open(fn, 'rt') as f:
			return f.read().splitlines()

	def iter_connected_monitors(self, test_input: typing.Optional[str] = None) -> typing.Iterator[model.Monitor]:
		monitor = None
		if test_input:
			xrandr_output = self._read_test_file(test_input)
		else:
			xrandr_output = self._get_xrandr_output()

		for ln in xrandr_output:
			if m := self.reo_connected.match(ln):
				if monitor:
					yield monitor

				monitor = model.Monitor(
					name = m.group('name'),
					long_name = '',
					on = m.group('width') is not None,
				)
			elif m := self.reo_available_resolution.match(ln):
				assert monitor
				if monitor.default_resolution is None or m.group('default'):
					monitor.default_resolution = model.Resolution(int(m.group('width')), int(m.group('height')))

		assert monitor
		yield monitor


	# ---------- write ----------

	direction_map = {
		model.Direction.LEFT : '--left-of',
		model.Direction.RIGHT : '--right-of',
		model.Direction.ABOVE : '--above',
		model.Direction.BELOW : '--below',
	}

	rotation_map = {
		model.Rotation.NORMAL : 'normal',
		model.Rotation.INVERTED : 'inverted',
		model.Rotation.LEFT : 'left',
		model.Rotation.RIGHT : 'right',
	}

	def turn_off_and_on(self,
			monitors_to_be_turned_off: typing.Iterable[model.Monitor],
			monitors_to_be_turned_on: typing.Iterable[model.Monitor],
			*, primary: typing.Optional[model.Monitor] = None) -> None:
		cmd = ['xrandr']
		for monitor in monitors_to_be_turned_off:
			cmd.extend(['--output', monitor.name, '--off'])
		for monitor in monitors_to_be_turned_on:
			cmd.extend(['--output', monitor.name])
			if monitor.position:
				cmd.append(self.direction_map[monitor.position.direction])
				cmd.append(monitor.position.reference_monitor.name)
			if monitor.same_as:
				cmd.append('--same-as')
				cmd.append(monitor.same_as.name)
			if monitor.resolution:
				cmd.append('--mode')
				cmd.append('{r.width}x{r.height}'.format(r=monitor.resolution))
			else:
				cmd.append('--auto')
			if monitor.scale:
				cmd.append('--scale')
				cmd.append(f'{monitor.scale.x}x{monitor.scale.y}')
			if monitor.rotation:
				cmd.append('--rotate')
				cmd.append(self.rotation_map[monitor.rotation])

		if primary:
			cmd.extend(['--output', primary.name, '--primary'])

		if self.verbose:
			print(shlex.join(cmd))
		if not self.dry_run:
			try:
				subprocess.run(cmd, check=True, text=True, stderr=subprocess.PIPE)
			except subprocess.CalledProcessError as e:
				m = re.match('xrandr: Configure crtc ([0-9]) failed', e.stderr)
				if m:
					crtc = m.group(1)
					if crtc == '0':
						crtc = '1'
					else:
						crtc = '0'
					#https://askubuntu.com/questions/136139/xrandr-configure-crtc-0-failed-when-trying-to-change-resolution-on-external-m
					cmd.extend(['--crtc',  '1'])
					if self.verbose:
						print(shlex.join(cmd))
					subprocess.run(cmd, check=True)
				else:
					print(e.stderr)



if __name__ == '__main__':
	api = Api(verbose=True, dry_run=False)
	for m in api.iter_connected_monitors():
		print(m)



#!../venv/bin/pytest

import pytest

from crandr import main
from test_without_config import *


@pytest.fixture
def use_swaymsg() -> None:
	main.DEFAULT_BACKEND = 'swaymsg'


class TestToggle:

	def test_toggle_from_default(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/default_eDP-1_DP-4.json'
		args = ['--dry-run', '--test-input', fn, 'toggle']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output \'"Ancor Communications Inc MX279 G8LMRS023461"\' enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output \'"Unknown 0x484E 0x00000000"\' disable',
			'swaymsg -- workspace 42',
		]

	def test_toggle_to_external(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'toggle']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output DP-4 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output eDP-1 disable',
			'swaymsg -- workspace 42',
		]

	def test_toggle_to_internal(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/DP-4_ON_eDP-1_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'toggle']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output eDP-1 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output DP-4 disable',
			'swaymsg -- workspace 42',
		]


class TestCycle:

	def test_cycle_from_default(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/default_eDP-1_DP-4.json'
		args = ['--dry-run', '--test-input', fn, 'cycle']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output \'"Unknown 0x484E 0x00000000"\' enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output \'"Ancor Communications Inc MX279 G8LMRS023461"\' disable',
			'swaymsg -- workspace 42',
		]

	def test_cycle_to_external(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'cycle']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output DP-4 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output eDP-1 disable',
			'swaymsg -- workspace 42',
		]

	def test_cycle_to_internal(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/DP-4_ON_eDP-1_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'cycle']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output eDP-1 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output DP-4 disable',
			'swaymsg -- workspace 42',
		]

	def test_cycle_prefer_external_both_enabled(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/internal_ON_external_ON.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'cycle', 'external,any']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output DP-3 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output eDP-1 disable',
			'swaymsg -- workspace 42',
		]

	def test_cycle_prefer_external_external_enabled(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/internal_OFF_external_ON.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'cycle', 'external,any']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output eDP-1 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output DP-3 disable',
			'swaymsg -- workspace 42',
		]


class TestReset:

	def test_reset_turn_off(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/default_eDP-1_DP-4.json'
		args = ['--dry-run', '--test-input', fn, 'reset']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output \'"Unknown 0x484E 0x00000000"\' enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output \'"Ancor Communications Inc MX279 G8LMRS023461"\' disable',
			'swaymsg -- workspace 42',
		]

	def test_reset_stay_on_same_monitor_1(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/DP-4_ON_eDP-1_OFF.json'
		args = ['--dry-run', '--test-input', fn, 'reset']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output \'"Ancor Communications Inc MX279 G8LMRS023461"\' enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- workspace 42',
		]

	def test_reset_stay_on_same_monitor_2(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'reset']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output eDP-1 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- workspace 42',
		]

	def test_reset_switch_monitor(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'reset', 'external']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output DP-4 enable position 0 0 scale 1.0 transform normal',
			'swaymsg -- output eDP-1 disable',
			'swaymsg -- workspace 42',
		]


class TestExtend:

	def test_extend_left(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, 'extend', 'left']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'swaymsg -- output \'"Ancor Communications Inc MX279 G8LMRS023461"\' enable position -1920 0 scale 1.0'

	def test_extend_right(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, 'extend', 'right']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'swaymsg -- output \'"Ancor Communications Inc MX279 G8LMRS023461"\' enable position 1600 0 scale 1.0'

	def test_extend_above(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'extend', 'above']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'swaymsg -- output DP-4 enable position 0 -1080 scale 1.0'

	def test_extend_below(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'extend', 'below']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'swaymsg -- output DP-4 enable position 0 900 scale 1.0'


class TestMirror:

	def test_mirror_one(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'mirror']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output eDP-1 enable position 0 0 mode 1600x900 scale 0.8333333333333334',
			'swaymsg -- output DP-4 enable position 0 0 mode 1920x1080 scale 1.0'
		]

	def test_mirror_default_to_highest_resolution(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/DP-3_ON_eDP-1_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'mirror']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip().splitlines()
		assert cmd == [
			'swaymsg -- output DP-3 enable position 0 0 mode 1920x1080 scale 1.0',
			'swaymsg -- output eDP-1 enable position 0 0 mode 1920x1080 scale 1.0',
		]


class TestScale:

	def test_scale_big(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, 'scale', '.8']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'swaymsg -- output \'"Unknown 0x484E 0x00000000"\' enable position 0 0 scale 1.25'

	def test_scale_reset(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'scale', '1']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'swaymsg -- output eDP-1 enable position 0 0 scale 1.0'

	def test_scale_different_x_and_y_not_possible(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, 'scale', '.8x1.2']
		
		with pytest.raises(SystemExit) as e:
			main.main(args)
		
		assert e.type == SystemExit
		assert e.value.code == 1
		
		out, err = capsys.readouterr()
		cmd = out.rstrip()
		err = err.rstrip()
		
		assert cmd == ''
		assert err == 'swaymsg does not support different scales for x and y direction'


class TestRotate:

	def test_rotate_left_internal_monitor(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, 'rotate', 'left']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'swaymsg -- output \'"Unknown 0x484E 0x00000000"\' enable position 0 0 transform 270'

	def test_rotate_inverted_internal_monitor(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'rotate', 'inverted']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'swaymsg -- output eDP-1 enable position 0 0 transform 180'

	def test_rotate_right_external_monitor(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/DP-4_ON_eDP-1_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'rotate', 'right']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'swaymsg -- output DP-4 enable position 0 0 transform 90'

	def test_rotate_normal_external_monitor(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/DP-4_ON_eDP-1_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'rotate', 'normal']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'swaymsg -- output DP-4 enable position 0 0 transform normal'


class TestList:

	def test_list(self, use_swaymsg: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/swaymsg/eDP-1_ON_DP-4_OFF.json'
		args = ['--dry-run', '--test-input', fn, '--use-short-names-on-sway', 'list']
		main.main(args)
		output = capsys.readouterr().out.rstrip()

		expected_output = '''\
eDP-1    (internal)      [on]   1600x900   "Unknown 0x484E 0x00000000"
DP-4     (display-port)  [off]  1920x1080  "Ancor Communications Inc MX279 G8LMRS023461"'''

		assert output == expected_output



#!/usr/bin/env python3

import os
import typing

class ConfigPathFinder:

	def __init__(self, app_name: str, *, file_name: str = 'config'):
		self.app_name = app_name
		self.file_name = file_name
		self._appdirs: typing.Any = None

	def get_app_dirs(self) -> typing.Any:
		if self._appdirs:
			return self._appdirs
		
		try:
			from platformdirs import PlatformDirs as AppDirs
		except:
			try:
				from xdgappdirs import AppDirs
			except:
				try:
					from appdirs import AppDirs
				except:
					return None
		self._appdirs = AppDirs(self.app_name, multipath=True)
		return self._appdirs

	def iter_config_directories(self) -> typing.Iterator[str]:
		app_dirs = self.get_app_dirs()
		if app_dirs:
			yield from app_dirs.user_config_dir.split(os.pathsep)
			yield from app_dirs.site_config_dir.split(os.pathsep)
		else:
			path = os.environ.get('XDG_CONFIG_HOME', None)
			if not path:
				path = os.path.expanduser('~')
				path = os.path.join(path, '.config')
			path = os.path.join(path, self.app_name)
			yield path

	def iter_possible_config_files(self) -> typing.Iterator[str]:
		for path in self.iter_config_directories():
			yield os.path.join(path, self.file_name)

	def get_existing_config_file(self) -> typing.Optional[str]:
		for ffn in self.iter_possible_config_files():
			if os.path.isfile(ffn):
				return ffn
		
		return None



#!../venv/bin/pytest

import os
import shutil
import tempfile
import typing
from types import TracebackType

import pytest

from crandr import main
from test_without_config import *


class configfile:

	def __init__(self, fn_in: str) -> None:
		self.fn_in = os.path.abspath(fn_in)

	def __enter__(self) -> str:
		self.path = tempfile.mkdtemp()
		self.old_config_home = os.environ.get('XDG_CONFIG_HOME', None)
		os.environ['XDG_CONFIG_HOME'] = self.path
		
		self.fn_out = os.path.join(self.path, main.APP_NAME, 'config')
		os.mkdir(os.path.join(self.path, main.APP_NAME))
		os.symlink(self.fn_in, self.fn_out)
		return self.fn_out

	def __exit__(self, exc_type: 'type[BaseException]', exc_val: 'BaseException', exc_tb: 'TracebackType') -> 'typing.Literal[False]':
		if self.old_config_home is None:
			del os.environ['XDG_CONFIG_HOME']
		else:
			os.environ['XDG_CONFIG_HOME'] = self.old_config_home
		
		shutil.rmtree(self.path)
		return False


class TestConfig:

	# ------- --list-searched-paths -------

	def test_config_list_searched_paths(self, capsys: 'pytest.CaptureFixture[str]') -> None:
		args = ['--dry-run', 'config', '--list-searched-paths']
		main.main(args)
		out = capsys.readouterr().out.rstrip()
		searched_path = out.splitlines()
		assert len(searched_path) >= 1
		
		for fn in searched_path:
			assert os.path.isabs(fn)
			assert os.path.split(fn)[1] == 'config'

	# ------- --print-file-name -------

	def test_config_print_nothing(self, capsys: 'pytest.CaptureFixture[str]') -> None:
		args = ['--dry-run', 'config', '--print-file-name']
		main.main(args)
		out = capsys.readouterr().out.rstrip()
		fn = out
		assert fn == ''

	def test_config_print_config_file(self, capsys: 'pytest.CaptureFixture[str]') -> None:
		args = ['--dry-run', 'config', '--print-file-name']
		with configfile('_test-output/config/test') as fn:
			main.main(args)
			out = capsys.readouterr().out.rstrip()
			assert out == fn

	# ------- --edit -------

	def test_config_edit__no_editor__without_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', '')
		args = ['--dry-run', 'config', '--edit']
		main.main(args)
		out = capsys.readouterr().out.rstrip().splitlines()
		path = os.path.join(os.environ['XDG_CONFIG_HOME'], main.APP_NAME)
		fn = os.path.join(path, 'config')
		assert out == [
			f'os.makedirs({path!r}, exist_ok=True)',
			f'cp \'doc/example_config\' {fn!r}',
			'vi %s' % fn,
		]

	def test_config_edit__no_editor__with_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', '')
		args = ['--dry-run', 'config', '--edit']
		with configfile('_test-output/config/test') as fn:
			main.main(args)
			out = capsys.readouterr().out.rstrip()
			assert out == 'vi %s' % fn

	def test_config_edit__with_editor__without_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', 'vim')
		args = ['--dry-run', 'config', '--edit']
		main.main(args)
		out = capsys.readouterr().out.rstrip().splitlines()
		path = os.path.join(os.environ['XDG_CONFIG_HOME'], main.APP_NAME)
		fn = os.path.join(path, 'config')
		assert out == [
			f'os.makedirs({path!r}, exist_ok=True)',
			f'cp \'doc/example_config\' {fn!r}',
			'vim %s' % fn,
		]

	def test_config_edit__with_editor__with_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', 'nano')
		args = ['--dry-run', 'config', '--edit']
		with configfile('_test-output/config/test') as fn:
			main.main(args)
			out = capsys.readouterr().out.rstrip()
			assert out == 'nano %s' % fn

	# ------- no arguments -------

	def test_config_default__no_editor__without_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', '')
		args = ['--dry-run', 'config']
		main.main(args)
		out = capsys.readouterr().out.rstrip().splitlines()
		assert out[:3] == [
			'[not opening config file because EDITOR is not defined]',
			'no config file existing',
			'the following paths are searched:'
		]
		assert len(out) >= 4
		assert out[3].startswith('- ')
		assert out[3].endswith('config')

	def test_config_default__no_editor__with_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', '')
		args = ['--dry-run', 'config']
		with configfile('_test-output/config/test') as fn:
			main.main(args)
			out = capsys.readouterr().out.rstrip().splitlines()
			assert out == [
				'[not opening config file because EDITOR is not defined]',
				fn,
			]

	def test_config_default__with_editor__without_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', 'nano')
		args = ['--dry-run', 'config']
		main.main(args)
		out = capsys.readouterr().out.rstrip().splitlines()
		path = os.path.join(os.environ['XDG_CONFIG_HOME'], main.APP_NAME)
		fn = os.path.join(path, 'config')
		assert out == [
			f'os.makedirs({path!r}, exist_ok=True)',
			f'cp \'doc/example_config\' {fn!r}',
			'nano %s' % fn,
		]

	def test_config_default__with_editor__with_config(self, capsys: 'pytest.CaptureFixture[str]', monkeypatch: pytest.MonkeyPatch) -> None:
		monkeypatch.setitem(os.environ, 'EDITOR', 'vim')
		args = ['--dry-run', 'config']
		with configfile('_test-output/config/test') as fn:
			main.main(args)
			out = capsys.readouterr().out.rstrip()
			assert out == 'vim %s' % fn


class TestConfigParsing:

	def test_map_name_to_connection(self, capsys: 'pytest.CaptureFixture[str]') -> None:
		args = ['--dry-run', '--backend', 'xrandr', '--test-input', '_test-output/xrandr/eDP-1_mirror_HDMI-1_original.txt', 'list', '--format', '{m.name}:{m.connection}']
		with configfile('_test-output/config/test') as fn:
			main.main(args)
			out = capsys.readouterr().out.rstrip().splitlines()
			assert out == [
				'eDP-1:internal',
				'HDMI-1:display-port',
			]



#!../venv/bin/pytest

import pytest

from crandr import main
from test_without_config import *


@pytest.fixture
def use_xrandr() -> None:
	main.DEFAULT_BACKEND = 'xrandr'


class TestDefault:

	def test_default_is_cycle(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_ON__DP-1-2_OFF.txt'
		
		args = ['--dry-run', '--test-input', fn]
		main.main(args)
		cmd_default = capsys.readouterr().out.rstrip()
		
		args.append('cycle')
		main.main(args)
		cmd_cycle = capsys.readouterr().out.rstrip()
		
		assert cmd_default == cmd_cycle


class TestToggle:

	def test_toggle_to_internal_from_dp11(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_ON__DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'toggle']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-1 --off --output eDP-1 --auto --scale 1x1 --rotate normal'

	def test_toggle_to_internal_from_dp12(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_OFF_DP-1-2_ON_.txt'
		args = ['--dry-run', '--test-input', fn, 'toggle']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-2 --off --output eDP-1 --auto --scale 1x1 --rotate normal'

	def test_toggle_to_external(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__DP-1-1_OFF_DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'toggle']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output eDP-1 --off --output DP-1-1 --auto --scale 1x1 --rotate normal'


class TestCycle:

	def test_cycle_to_dp11(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__DP-1-1_OFF_DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'cycle']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output eDP-1 --off --output DP-1-1 --auto --scale 1x1 --rotate normal'

	def test_cycle_to_dp12(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_ON__DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'cycle']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-1 --off --output DP-1-2 --auto --scale 1x1 --rotate normal'

	def test_cycle_to_edp1(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_OFF_DP-1-2_ON_.txt'
		args = ['--dry-run', '--test-input', fn, 'cycle']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-2 --off --output eDP-1 --auto --scale 1x1 --rotate normal'


class TestReset:

	def test_reset_dont_switch_monitor(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_OFF_DP-1-2_ON_.txt'
		args = ['--dry-run', '--test-input', fn, 'reset']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-2 --auto --scale 1x1 --rotate normal'

	def test_reset_turn_off(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_mirror_HDMI-1_original.txt'
		args = ['--dry-run', '--test-input', fn, 'reset']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output HDMI-1 --off --output eDP-1 --auto --scale 1x1 --rotate normal'

	def test_reset_turn_other_off(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_mirror_HDMI-1_original.txt'
		args = ['--dry-run', '--test-input', fn, 'reset', 'external']
		main.main(args)
		
		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output eDP-1 --off --output HDMI-1 --auto --scale 1x1 --rotate normal'


class TestExtend:

	def test_extend_left(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_ON__DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'extend', 'left']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-2 --left-of DP-1-1 --auto --scale 1x1 --output DP-1-1 --primary'

	def test_extend_left_internal(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_ON__DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'extend', 'left', 'internal']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output eDP-1 --left-of DP-1-1 --auto --scale 1x1 --output DP-1-1 --primary'

	def test_extend_right(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_ON__DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'extend', 'right', 'external']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-2 --right-of DP-1-1 --auto --scale 1x1 --output DP-1-1 --primary'

	def test_extend_above(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__DP-1-1_OFF_DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'extend', 'above']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-1 --above eDP-1 --auto --scale 1x1 --output eDP-1 --primary'

	def test_extend_below(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__DP-1-1_OFF_DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'extend', 'below', 'external']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output DP-1-1 --below eDP-1 --auto --scale 1x1 --output eDP-1 --primary'


class TestMirror:

	def test_mirror_one(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__HDMI-1_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'mirror']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output eDP-1 --same-as HDMI-1 --mode 1600x900 --scale 1.2x1.2 --output HDMI-1 --mode 1920x1080 --scale 1x1'

	def test_mirror_change_original(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_mirror_HDMI-1_original.txt'
		args = ['--dry-run', '--test-input', fn, 'mirror', '--original', 'internal']
		main.main(args)

		cmd = capsys.readouterr().out.rstrip()
		assert cmd == 'xrandr --output eDP-1 --mode 1600x900 --scale 1x1 --output HDMI-1 --same-as eDP-1 --mode 1920x1080 --scale 0.8333333333333334x0.8333333333333334'


class TestScale:

	def test_scale_wide(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__HDMI-1_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'scale', '.8x1.2']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'xrandr --output eDP-1 --auto --scale 0.8x1.2'

	def test_scale_reset(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__HDMI-1_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'scale', '1']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'xrandr --output eDP-1 --auto --scale 1.0x1.0'


class TestRotate:

	def test_rotate_left(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_ON__HDMI-1_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'rotate', 'left']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'xrandr --output eDP-1 --auto --rotate left'

	def test_rotate_invert_two_monitors(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_left-of_DP-1-2.txt'
		args = ['--dry-run', '--test-input', fn, 'rotate', 'inverted']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'xrandr --output eDP-1 --auto --rotate inverted --output DP-1-2 --auto --rotate inverted'

	def test_rotate_right_external_monitor(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_left-of_DP-1-2.txt'
		args = ['--dry-run', '--test-input', fn, 'rotate', 'right', 'external']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'xrandr --output DP-1-2 --auto --rotate right'

	def test_rotate_normal_internal_monitor(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_left-of_DP-1-2.txt'
		args = ['--dry-run', '--test-input', fn, 'rotate', 'normal', 'internal']
		
		main.main(args)
		cmd = capsys.readouterr().out.rstrip()
		
		assert cmd == 'xrandr --output eDP-1 --auto --rotate normal'


class TestList:

	def test_list(self, use_xrandr: None, capsys: 'pytest.CaptureFixture[str]') -> None:
		fn = '_test-output/xrandr/eDP-1_OFF_DP-1-1_ON__DP-1-2_OFF.txt'
		args = ['--dry-run', '--test-input', fn, 'list']
		main.main(args)
		output = capsys.readouterr().out.rstrip()

		expected_output = '''\
eDP-1    (internal)      [off]  1600x900
DP-1-1   (display-port)  [on]   1920x1080
DP-1-2   (display-port)  [off]  1920x1080'''

		assert output == expected_output



#!../venv/bin/pytest

import pytest

from crandr import utils

@pytest.fixture(scope='module')
def formatter() -> 'utils.MyFormatter':
	return utils.MyFormatter()

class TestFormatterStandardFeatures:

	def test_format_int(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('that is {0}', 1)
		assert result == 'that is 1'

	def test_format_floats(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('{0:.2} != {1:.2}', 1/2, 1/3)
		assert result == '0.5 != 0.33'

	def test_format_str(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('hello {addressee}', addressee='world')
		assert result == 'hello world'

	def test_format_str_repr(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('a = {a!r}', a='')
		assert result == "a = ''"

	def test_format_empty(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('', foo=42)
		assert result == ''

	def test_format_list(self, formatter: 'utils.MyFormatter') -> None:
		l = ['hi', 'there']
		result = formatter.format('here and {l[1]}', l=l)
		assert result == 'here and there'

	def test_format_padding(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('|{0:>10}|{1:<10}|', 'hello', 'world')
		assert result == '|     hello|world     |'


class TestFormatterExtendedFeatures:

	def test_format_floats_in_parentheses(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('{"("0")":5.2f} != {"("1")":5.2f}', 1/2, 1/3)
		assert result == ' (0.50) !=  (0.33)'

	def test_format_prefix_and_suffix_are_omitted_if_value_is_empty(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('--{"[ "0" ]"}--', '')
		assert result == '----'

	def test_format_str_repr_with_prefix(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('a{" = "a!r}', a='')
		assert result == "a = ''"

	def test_format_example_1(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('{name:8}  {"["on"]":3}  {resolution}', name='DP-1', on='on', resolution='1920x1080')
		assert result == 'DP-1      [on]   1920x1080'

	def test_format_example_2(self, formatter: 'utils.MyFormatter') -> None:
		result = formatter.format('{name:8}  {"["on"]":3}  {resolution}', name='HDMI-2', on='off', resolution='1920x1080')
		assert result == 'HDMI-2    [off]  1920x1080'

	def test_format_default(self, formatter: 'utils.MyFormatter') -> None:
		class Monitor:
			name = 'eDP-1'
			connection = 'internal'
			on_str = 'on'
			default_resolution = '1600x900'
		m = Monitor
		pattern = '{m.name:8} {"("m.connection")":12}  {"["m.on_str"]":3}  {m.default_resolution}'
		result = formatter.format(pattern, m=m)
		assert result == 'eDP-1    (internal)      [on]   1600x900'




#!/usr/bin/env python3

from __future__ import annotations

import fnmatch
import enum
import typing

CONNECTION_INTERNAL = 'internal'
CONNECTION_DP = 'display-port'
CONNECTION_HDMI = 'hdmi'
CONNECTION_VGA = 'vga'
name_to_connection: typing.MutableMapping[str, str] = {}
EXTERNAL_CONNECTIONS = (CONNECTION_DP, CONNECTION_HDMI, CONNECTION_VGA)
ALL_CONNECTIONS = (CONNECTION_INTERNAL,) + EXTERNAL_CONNECTIONS


class Monitor:

	__slots__ = ('name', 'long_name', 'on', 'position', 'default_resolution', 'same_as', 'resolution', 'scale', 'rotation')

	@property
	def connection(self) -> str:
		for name in sorted(name_to_connection, key=len, reverse=True):
			if fnmatch.fnmatchcase(self.name, name):
				return name_to_connection[name]
		if self.name.startswith("DP"):
			return CONNECTION_DP
		if self.name.startswith("HDMI"):
			return CONNECTION_HDMI
		if self.name.startswith("VGA"):
			return CONNECTION_VGA
		else:
			return CONNECTION_INTERNAL

	def __init__(self, *, name: str, long_name:str, on: bool,
			default_resolution: typing.Optional['Resolution'] = None,
			position: typing.Optional['Position'] = None) -> None:
		# read
		self.name = name
		self.long_name = long_name
		self.on = on
		self.default_resolution = default_resolution

		# write
		self.position = position
		self.same_as: typing.Optional[Monitor] = None
		self.resolution: typing.Optional[Resolution] = None
		self.scale: typing.Optional[Scale] = None
		self.rotation: typing.Optional[Rotation] = None

	def __repr__(self) -> str:
		attrs = ', '.join('%s=%r' % (attr, getattr(self, attr)) for attr in self.__slots__)
		return "%s(%s)" % (type(self).__name__, attrs)

	@property
	def on_str(self) -> str:
		if self.on:
			return "on"
		else:
			return "off"

def sort_monitors_by_connection(monitors: 'list[Monitor]', connections: 'list[str]') -> None:
	n = len(connections)
	monitors.sort(key = lambda m: find_in_list(connections, m.connection, n))

T = typing.TypeVar('T')
def find_in_list(l: 'list[T]', obj: 'T', default: int = -1) -> int:
	try:
		return l.index(obj)
	except ValueError:
		return default


@enum.unique
class Direction(enum.Enum):

	LEFT = 'left'
	RIGHT = 'right'
	ABOVE = 'above'
	BELOW = 'below'

class Position:

	__slots__ = ('direction', 'reference_monitor')

	def __init__(self, direction: Direction, reference_monitor: Monitor):
		self.direction = direction
		self.reference_monitor = reference_monitor


class Resolution:

	__slots__ = ('width', 'height')

	def __init__(self, width: int, height: int):
		self.width = width
		self.height = height

	def __truediv__(self, other: Resolution) -> Scale:
		if not isinstance(other, Resolution):
			return NotImplemented
		return Scale(
			self.width / other.width,
			self.height / other.height,
		)

	def __str__(self) -> str:
		return f'{self.width}x{self.height}'

	def __repr__(self) -> str:
		return f"{type(self).__name__}({self.width}, {self.height})"

class Scale:

	__slots__ = ('x', 'y')

	def __init__(self, x: float, y: float):
		self.x = x
		self.y = y

	@classmethod
	def parse(cls, s: str) -> Scale:
		if 'x' in s:
			sx, sy = s.split('x', 1)
			x = float(sx)
			y = float(sy)
		else:
			x = float(s)
			y = x

		return cls(x, y)


@enum.unique
class Rotation(enum.Enum):

	NORMAL = 'normal'
	INVERTED = 'inverted'
	LEFT = 'left'
	RIGHT = 'right'


class NotSupportedException(Exception):
	 pass


class Api:

	def __init__(self, *, verbose: bool, dry_run: bool, **kw: typing.Mapping[str, typing.Any]) -> None:
		self.verbose = verbose
		self.dry_run = dry_run
		pass

	def print_version(self) -> None:
		raise NotImplementedError()

	def iter_connected_monitors(self, test_input: typing.Optional[str] = None) -> typing.Iterator[Monitor]:
		raise NotImplementedError()

	def turn_off_and_on(self,
			monitors_to_be_turned_off: typing.Iterable[Monitor],
			monitors_to_be_turned_on: typing.Iterable[Monitor],
			*, primary: typing.Optional[Monitor] = None) -> None:
		raise NotImplementedError()


if __name__ == '__main__':
	pass



#!/usr/bin/env python3

import string
import re
import typing

class MyFormatter(string.Formatter):

	"""
	This should be fully compatible with the default Formatter but adds one feature:
	A prefix and suffix can be specified inside of the field name, wrapped in single or double quotes.
	If the resulting value is not empty the prefix is inserted after leading white space and
	the suffix is inserted before trailing white space.
	Prefix and suffix do not count in to the padding length.
	
	>>> MyFormatter().format('{name:8}  {"["on"]":3}  {resolution}', name='DP-1', on='on', resolution='1920x1080')
	'DP-1      [on]   1920x1080'
	>>> MyFormatter().format('{name:8}  {"["on"]":3}  {resolution}', name='HDMI-2', on='off', resolution='1920x1080')
	'HDMI-2    [off]  1920x1080'
	"""

	QUOTES = ('"', "'")

	def get_field(self, field_name: str, args: typing.Sequence[typing.Any], kwargs: typing.Mapping[str, typing.Any]) -> typing.Tuple[typing.Tuple[str, typing.Any, str], str]:
		prefix = ""
		suffix = ""
		if field_name:
			if field_name[0] in self.QUOTES:
				quote = field_name[0]
				i = field_name.index(quote, 1)
				prefix = field_name[1:i]
				field_name = field_name[i+1:]
			if field_name[-1] in self.QUOTES:
				quote = field_name[-1]
				i = field_name.rindex(quote, 0, -1)
				suffix = field_name[i+1:-1]
				field_name = field_name[:i]
		
		value, used_key = super().get_field(field_name, args, kwargs)
		return (prefix, value, suffix), used_key

	def convert_field(self, value_tuple: typing.Tuple[str, typing.Any, str], conversion: typing.Optional[str]) -> typing.Tuple[str, str, str]:
		prefix, value, suffix = value_tuple
		formatted_value = super().convert_field(value, conversion)
		return prefix, formatted_value, suffix

	def format_field(self, value_tuple: typing.Tuple[str, typing.Any, str], format_spec: str) -> typing.Any:
		prefix, value, suffix = value_tuple
		formatted_value = super().format_field(value, format_spec)
		if value:
			m = re.match(r'(\s*)(.*?)(\s*)$', formatted_value)
			assert m
			formatted_value = m.group(1) + prefix + m.group(2) + suffix + m.group(3)
		return formatted_value


if __name__ == '__main__':
	class Monitor:
		on_str = 'on'
		name = 'HDMI-1'
		connection = 'vga'
		default_resolution = "1920x1080"

	pattern = '{m.name:8} {"("m.connection")":12}  {"["m.on_str"]"!s:3}  {m.default_resolution}'
	out = MyFormatter().format(pattern, m=Monitor())
	print(out)



#!../venv/bin/pytest

import os
import tempfile
from collections.abc import Iterator

import pytest

@pytest.fixture(autouse=True)
def no_config_file(monkeypatch: pytest.MonkeyPatch) -> 'Iterator[None]':
	path = tempfile.mkdtemp()
	monkeypatch.setitem(os.environ, 'XDG_CONFIG_HOME', path)
	yield
	os.rmdir(path)



"""
This is a unifying wrapper around xrandr and swaymsg to control your monitors/outputs,
mainly intended for setups where only one or two monitors are enabled at the same time.
"""

from .version import __version__



#!../venv/bin/pytest

from crandr.model import Monitor, Resolution, sort_monitors_by_connection
from crandr.connection_parser import connection

def test_sort_monitors_by_connection__prefer_external() -> None:
	monitors = [
		Monitor(name="eDP-1", long_name="", on=True),
		Monitor(name="DP-3", long_name="", on=True),
	]
	connections = connection("external,any")
	sort_monitors_by_connection(monitors, connections)
	assert monitors[0].name == "DP-3"
	assert monitors[1].name == "eDP-1"
	assert len(monitors) == 2

def test_sort_monitors_by_connection__prefer_internal() -> None:
	monitors = [
		Monitor(name="eDP-1", long_name="", on=True),
		Monitor(name="DP-3", long_name="", on=True),
	]
	connections = connection("internal,any")
	sort_monitors_by_connection(monitors, connections)
	assert monitors[0].name == "eDP-1"
	assert monitors[1].name == "DP-3"
	assert len(monitors) == 2

def test_repr_resolution() -> None:
	r = Resolution(1920, 1080)
	r2 = eval(repr(r), {}, {'Resolution': Resolution})
	assert type(r) is type(r2)
	assert r.width == r2.width
	assert r.height == r.height



#!/usr/bin/env python3

import subprocess
import json
import typing
import shlex

from . import model


def is_available() -> bool:
	cmd = ['which', 'swaymsg']
	try:
		subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		return True
	except:
		return False


class Api(model.Api):

	encoding = 'utf-8'

	def __init__(self, *, verbose: bool, dry_run: bool, **kw: typing.Mapping[str, typing.Any]) -> None:
		super().__init__(verbose=verbose, dry_run=dry_run)
		self.use_short_names = typing.cast(bool, kw.get('use_short_names_on_sway', False))


	# ---------- meta ----------

	def print_version(self) -> None:
		subprocess.call(['swaymsg', '--version'])


	# ---------- read ----------

	def _get_outputs(self) -> str:
		cmd = ['swaymsg', '-t', 'get_outputs']
		p = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
		return p.stdout.decode(self.encoding)

	def _read_test_file(self, fn: str) -> str:
		with open(fn, 'rt') as f:
			return f.read()

	def iter_connected_monitors(self, test_input: typing.Optional[str] = None) -> typing.Iterator[model.Monitor]:
		monitor = None
		if test_input:
			outputs_str = self._read_test_file(test_input)
		else:
			outputs_str = self._get_outputs()

		outputs = json.loads(outputs_str)
		for o in outputs:
			default_mode = max(o['modes'], key=lambda m: m['width'] * m['height'])
			yield model.Monitor(
				name = o['name'],
				long_name = '"%s %s %s"' % (o['make'], o['model'], o['serial']),
				on = o['active'],
				default_resolution = model.Resolution(
					default_mode['width'],
					default_mode['height'],
				)
			)

	def get_current_workspace(self, test_input: typing.Optional[str] = None) -> str:
		if test_input:
			return '42'

		cmd = ['swaymsg', '-r', '-t', 'get_workspaces']
		p = subprocess.run(cmd, stdout=subprocess.PIPE, check=True)
		workspaces = p.stdout.decode(self.encoding)

		for ws in json.loads(workspaces):
			if ws.get('type') == 'workspace' and ws.get('focused'):
				name = ws['name']
				assert isinstance(name, str)
				return name

		assert False, "no focused workspace found"


	# ---------- write ----------

	rotation_map = {
		model.Rotation.NORMAL : 'normal',
		model.Rotation.RIGHT : '90',
		model.Rotation.INVERTED : '180',
		model.Rotation.LEFT : '270',
	}

	def turn_off_and_on(self,
			monitors_to_be_turned_off: typing.Iterable[model.Monitor],
			monitors_to_be_turned_on: typing.Iterable[model.Monitor],
			*, primary: typing.Optional[model.Monitor] = None) -> None:
		get_name: typing.Callable[[model.Monitor], str]
		if self.use_short_names:
			get_name = lambda m: m.name
		else:
			get_name = lambda m: m.long_name
		for monitor in monitors_to_be_turned_on:
			x, y = self._calc_position(monitor)
			cmd = ['swaymsg', '--', 'output', get_name(monitor), 'enable', 'position', f'{x}', f'{y}']
			if monitor.resolution:
				cmd.append('mode')
				cmd.append('{r.width}x{r.height}'.format(r=monitor.resolution))
			if monitor.scale:
				if monitor.scale.x != monitor.scale.y:
					raise model.NotSupportedException('swaymsg does not support different scales for x and y direction')
				scale = 1 / monitor.scale.x
				cmd.append('scale')
				cmd.append(f'{scale}')
			if monitor.rotation:
				cmd.append('transform')
				cmd.append(self.rotation_map[monitor.rotation])
			self._run_swaymsg(cmd)
		for monitor in monitors_to_be_turned_off:
			cmd = ['swaymsg', '--', 'output', get_name(monitor), 'disable']
			self._run_swaymsg(cmd)

	def _calc_position(self, monitor: model.Monitor) -> typing.Tuple[int, int]:
		#TODO: consider scale
		def get_resolution(m: model.Monitor) -> model.Resolution:
			if m.resolution:
				return m.resolution
			else:
				assert m.default_resolution
				return m.default_resolution

		# the position specifies the top left corner
		# the y axis is pointing downward
		if monitor.position is None:
			x = 0
			y = 0
		elif monitor.position.direction == model.Direction.LEFT:
			x = - get_resolution(monitor).width
			y = 0
		elif monitor.position.direction == model.Direction.RIGHT:
			x = get_resolution(monitor.position.reference_monitor).width
			y = 0
		elif monitor.position.direction == model.Direction.ABOVE:
			x = 0
			y = - get_resolution(monitor).height
		elif monitor.position.direction == model.Direction.BELOW:
			x = 0
			y = get_resolution(monitor.position.reference_monitor).height
		else:
			assert False

		return x, y

	def set_current_workspace(self, name: str) -> None:
		self._run_swaymsg(['swaymsg', '--', 'workspace', name])

	def _run_swaymsg(self, cmd: typing.Sequence[str]) -> None:
		if self.verbose:
			print(shlex.join(cmd))
		if not self.dry_run:
			subprocess.run(cmd, check=True)


if __name__ == '__main__':
	api = Api(verbose=True, dry_run=False)
	for m in api.iter_connected_monitors():
		print(m)



#!/usr/bin/env python3

# Copyright © 2022 erzo <erzo@posteo.de>
# This work is free. You can use, copy, modify, and/or distribute it
# under the terms of the BSD Zero Clause License, see LICENSE.

"""
This is a unifying wrapper around xrandr and swaymsg to control your monitors/outputs,
mainly intended for setups where only one or two monitors are enabled at the same time.

All sub commands take an optional connections argument to specify which monitor you want to use.
Unfortunately it's value is not always as expected. For example xrandr reports my VGA port as DP-2
and one of the display ports as HDMI-1.
You can check the connection name to use with list, it is given in parentheses after the name.
You can manually set a connection name with `crandr config`.
"""

import os
import sys
import time
import argparse
import subprocess
import shlex
import typing
import importlib
import pkgutil
import configparser
from collections.abc import Iterable

from . import model
from . import api_swaymsg
from . import utils
from . import config_paths
from .connection_parser import connection, possible_connections, CONNECTION_ANY, CONNECTION_EXTERNAL
from .version import __version__

APP_NAME = 'crandr'

BACKENDS = {
	'swaymsg' : '.api_swaymsg',
	'xrandr' : '.api_xrandr',
	'auto' : None,
}
DEFAULT_BACKEND = 'auto'


# ---------- parse config file ----------

def parse_config() -> None:
	SECTION_MAP_NAME_TO_CONNECTION = 'map-name-to-connection'
	ALLOWED_SECTIONS = [SECTION_MAP_NAME_TO_CONNECTION]
	
	fn = config_paths.ConfigPathFinder(APP_NAME).get_existing_config_file()
	if not fn:
		return None
	
	class MyConfigParser(configparser.ConfigParser):
		# avoid conversion to lower case
		def optionxform(self, x: str) -> str:
			return x
	
	parser = MyConfigParser(
		allow_no_value = False,
		strict = True,
		empty_lines_in_values = False,
		interpolation = None,
	)
	
	try:
		parser.read(fn)
	except Exception as e:
		config_error(fn, e)
	
	for section in parser.sections():
		if section not in ALLOWED_SECTIONS:
			config_error(fn, 'unknown section %r, should be one of %s' % (section, ', '.join('%r' % s for s in ALLOWED_SECTIONS)))
	
	if SECTION_MAP_NAME_TO_CONNECTION in parser:
		mappings = parser[SECTION_MAP_NAME_TO_CONNECTION]
		for name, connection in mappings.items():
			if connection in model.ALL_CONNECTIONS:
				model.name_to_connection[name] = connection
			else:
				config_error(fn, 'invalid connection %r, should be one of %s' % (connection, ', '.join('%r' % c for c in model.ALL_CONNECTIONS)))

def config_error(fn: str, error: typing.Union[str, Exception]) -> None:
	print('error in config file %s:' % fn, file=sys.stderr)
	print(error, file=sys.stderr)
	sys.exit(2)


# ---------- enable one monitor ----------

def default(api: model.Api, args: argparse.Namespace) -> None:
	args.connections = model.ALL_CONNECTIONS
	cycle(api, args)

def cycle(api: model.Api, args: argparse.Namespace) -> None:
	'''cycle through all connected monitors, turning on one at a time'''
	if isinstance(api, api_swaymsg.Api):
		current_workspace = api.get_current_workspace(args.test_input)

	monitors = list(api.iter_connected_monitors(test_input=args.test_input))
	model.sort_monitors_by_connection(monitors, args.connections)
	if not monitors:
		print("no monitors connected", file=sys.stderr)
		sys.exit(1)

	monitors_to_be_turned_off = []
	monitor_to_be_turned_on = None

	was_last_on = False
	for m in monitors:
		if m.on:
			monitors_to_be_turned_off.append(m)
			was_last_on = True
		elif was_last_on and m.connection in args.connections:
			monitor_to_be_turned_on = m
			was_last_on = False

	if monitor_to_be_turned_on is None:
		for m in monitors:
			if m.connection in args.connections:
				monitor_to_be_turned_on = m
				break
		else:
			monitor_to_be_turned_on = get_fallback_monitor(monitors, args.connections)
		if monitor_to_be_turned_on in monitors_to_be_turned_off:
			monitors_to_be_turned_off.remove(monitor_to_be_turned_on)

	monitor_to_be_turned_on.scale = model.Scale(1, 1)
	monitor_to_be_turned_on.rotation = model.Rotation.NORMAL
	monitors_to_be_turned_on = [monitor_to_be_turned_on]
	api.turn_off_and_on(monitors_to_be_turned_off, monitors_to_be_turned_on)

	if isinstance(api, api_swaymsg.Api):
		time.sleep(args.time_in_s)
		api.set_current_workspace(current_workspace)

def toggle(api: model.Api, args: argparse.Namespace) -> None:
	'''toggle between two connections'''
	if isinstance(api, api_swaymsg.Api):
		current_workspace = api.get_current_workspace(args.test_input)

	monitors = list(api.iter_connected_monitors(test_input=args.test_input))
	if not monitors:
		print("no monitors connected", file=sys.stderr)
		sys.exit(1)
	
	monitors_to_be_turned_off = [m for m in monitors if m.on]
	
	if monitors_to_be_turned_off and monitors_to_be_turned_off[0].connection in args.connections1:
		connections = args.connections2
	else:
		connections = args.connections1

	for m in monitors:
		if not m.on and m.connection in connections:
			monitor_to_be_turned_on = m
			break
	else:
		alternative_connections = args.connections1 if connections is args.connections2 else args.connections2
		monitor_to_be_turned_on = get_fallback_monitor(monitors, connections, alternative_connections)
		if monitor_to_be_turned_on in monitors_to_be_turned_off:
			monitors_to_be_turned_off.remove(monitor_to_be_turned_on)
	
	monitor_to_be_turned_on.scale = model.Scale(1, 1)
	monitors_to_be_turned_on = [monitor_to_be_turned_on]
	monitor_to_be_turned_on.rotation = model.Rotation.NORMAL
	api.turn_off_and_on(monitors_to_be_turned_off, monitors_to_be_turned_on)

	if isinstance(api, api_swaymsg.Api):
		time.sleep(args.time_in_s)
		api.set_current_workspace(current_workspace)

def reset(api: model.Api, args: argparse.Namespace) -> None:
	'''turn off other monitors after extend or mirror and reset scaling and rotation after rotate, scale and mirror'''
	if isinstance(api, api_swaymsg.Api):
		current_workspace = api.get_current_workspace(args.test_input)

	monitors = list(api.iter_connected_monitors(test_input=args.test_input))
	if not monitors:
		print("no monitors connected", file=sys.stderr)
		sys.exit(1)
	
	monitor_to_be_on = get_fallback_monitor(monitors, args.connections)
	monitor_to_be_on.scale = model.Scale(1, 1)
	monitor_to_be_on.rotation = model.Rotation.NORMAL
	
	monitors_to_be_turned_off = [m for m in monitors if m.on and m is not monitor_to_be_on]
	monitors_to_be_turned_on = [monitor_to_be_on]
	api.turn_off_and_on(monitors_to_be_turned_off, monitors_to_be_turned_on)

	if isinstance(api, api_swaymsg.Api):
		time.sleep(args.time_in_s)
		api.set_current_workspace(current_workspace)

def get_fallback_monitor(monitors: typing.List[model.Monitor], connections: typing.Container[str], alternative_connections: typing.Container[str] = []) -> model.Monitor:
	monitors.sort(key=lambda m: 4*(m.connection in connections) + 2*(m.connection in alternative_connections) + m.on, reverse=True)
	return monitors[0]


# ---------- enable a second monitor ----------

def extend(api: model.Api, args: argparse.Namespace) -> None:
	'''turn on a second monitor'''
	monitors = list(api.iter_connected_monitors(test_input=args.test_input))
	model.sort_monitors_by_connection(monitors, args.connections)
	if not monitors:
		print("no monitors connected", file=sys.stderr)
		sys.exit(1)
	
	direction: model.Direction = args.direction
	try:
		reference_monitor = next(m for m in monitors if m.on)
	except StopIteration:
		print("no monitor turned on, I don't know which monitor to use as reference", file=sys.stderr)
		sys.exit(1)
	
	monitors_to_be_turned_off: typing.Final[typing.List[model.Monitor]] = []
	monitors_to_be_turned_on = [m for m in monitors if not m.on and m.connection in args.connections]
	if not monitors_to_be_turned_on:
		print("no further monitor to be turned on in %s" % ",".join(args.connections))
		sys.exit(1)
	
	monitors_to_be_turned_on = monitors_to_be_turned_on[:1]
	monitors_to_be_turned_on[0].position = model.Position(direction, reference_monitor)
	monitors_to_be_turned_on[0].scale = model.Scale(1, 1)
	
	api.turn_off_and_on(monitors_to_be_turned_off, monitors_to_be_turned_on, primary=reference_monitor)

def mirror(api: model.Api, args: argparse.Namespace) -> None:
	'''turn on all monitors and mirror the content'''
	monitors = list(api.iter_connected_monitors(test_input=args.test_input))
	model.sort_monitors_by_connection(monitors, args.connections)
	if not monitors:
		print("no monitors connected", file=sys.stderr)
		sys.exit(1)
	
	monitors_to_be_turned_off: typing.Final[typing.List[model.Monitor]] = []
	monitors_to_be_turned_on = [m for m in monitors if m.on]
	for m in monitors:
		if not m.on and m.connection in args.connections:
			monitors_to_be_turned_on.append(m)
	
	if len(monitors_to_be_turned_on) < 2:
		print("no second monitor to turn on", file=sys.stderr)
		sys.exit(1)
	
	try:
		original = next(m for m in monitors_to_be_turned_on if m.connection in args.original)
	except StopIteration:
		print("monitor chosen as original does not exist or is not active", file=sys.stderr)
		sys.exit(1)
	
	original.resolution = original.default_resolution
	assert original.resolution
	for m in monitors_to_be_turned_on:
		if m is not original:
			m.resolution = m.default_resolution
			assert m.resolution
			m.same_as = original
			m.scale = original.resolution / m.resolution
		else:
			m.scale = model.Scale(1, 1)
	
	api.turn_off_and_on(monitors_to_be_turned_off, monitors_to_be_turned_on)
	
	if api.__module__.endswith('api_swaymsg'):
		print('mirroring is not yet implemented in sway [February 2022]', file=sys.stderr)
		print('but I have set things up so that floating windows appear on both monitors', file=sys.stderr)
		print('using wayvnc + vinagre might be a better option, though', file=sys.stderr)
		print('https://github.com/swaywm/sway/issues/1666', file=sys.stderr)


# ---------- tweak monitors which are already turned on ----------

def rotate(api: model.Api, args: argparse.Namespace) -> None:
	'''rotate the monitor which is currently turned on'''
	monitors_to_be_rotated = [m for m in api.iter_connected_monitors(test_input=args.test_input) if m.on and m.connection in args.connections]
	if not monitors_to_be_rotated:
		print("no monitors to be rotated", file=sys.stderr)
		sys.exit(1)
	
	for m in monitors_to_be_rotated:
		m.rotation = args.rotation
	
	api.turn_off_and_on([], monitors_to_be_rotated)

def scale(api: model.Api, args: argparse.Namespace) -> None:
	'''scale the monitor which is currently turned on'''
	monitors_to_be_scaled = [m for m in api.iter_connected_monitors(test_input=args.test_input) if m.on and m.connection in args.connections]
	if not monitors_to_be_scaled:
		print("no monitors to be scaled", file=sys.stderr)
		sys.exit(1)
	
	for m in monitors_to_be_scaled:
		m.scale = args.scale
	
	api.turn_off_and_on([], monitors_to_be_scaled)


# ---------- list monitors ----------

def ls(api: model.Api, args: argparse.Namespace) -> None:
	'''list all connected monitors'''
	formatter = utils.MyFormatter()
	for m in api.iter_connected_monitors(test_input=args.test_input):
		if m.connection in args.connections:
			print(formatter.format(args.format, m=m).rstrip())


# ---------- config ----------

def config(api: model.Api, args: argparse.Namespace) -> None:
	config_finder = config_paths.ConfigPathFinder(APP_NAME)
	if args.list_searched_paths:
		for searched_path in config_finder.iter_possible_config_files():
			print(searched_path)
	elif args.print_file_name:
		fn = config_finder.get_existing_config_file()
		if fn:
			print(fn)
	else:
		editor = os.environ.get('EDITOR', None)
		if not editor:
			if args.edit:
				editor = 'vi'
			else:
				config_fallback(config_finder)
				return
		fn = config_finder.get_existing_config_file()
		if not fn:
			fn = next(config_finder.iter_possible_config_files())
			path = os.path.split(fn)[0]
			resource_example_config = 'doc/example_config'
			if args.verbose:
				print(f'os.makedirs({path!r}, exist_ok=True)')
				print(f'cp {resource_example_config!r} {fn!r}')
			if not args.dry_run:
				os.makedirs(path, exist_ok=True)
				with open(fn, 'wt') as f:
					f.write(get_resource_content(resource_example_config))
		cmd = editor + ' ' + shlex.quote(fn)
		if args.verbose:
			print(cmd)
		if not args.dry_run:
			subprocess.run(cmd, shell=True)
			parse_config()

def config_fallback(config_finder: config_paths.ConfigPathFinder) -> None:
	print('[not opening config file because EDITOR is not defined]')
	fn = config_finder.get_existing_config_file()
	if fn:
		print(fn)
	else:
		print('no config file existing')
		print('the following paths are searched:')
		for searched_path in config_finder.iter_possible_config_files():
			print('- %s' % searched_path)

def get_resource_content(resource_name: str) -> str:
	#https://stackoverflow.com/a/58941536
	raw = pkgutil.get_data(__name__, resource_name)
	assert raw is not None
	return raw.decode('utf-8')


# ---------- version ----------

def print_version(args: argparse.Namespace) -> None:
	'''show the version of this program and the used backend and exit'''
	print(f'{APP_NAME} {__version__}')
	# I am deliberately calling get_api only after printing the version number of this program in case get_api fails
	get_api(args).print_version()
	sys.exit(0)


# ---------- main ----------

def create_parser() -> argparse.ArgumentParser:
	DEFAULT_TIME_IN_S = 0.3

	root_parser = argparse.ArgumentParser(APP_NAME, description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
	subparsers = root_parser.add_subparsers(title='command')

	def add_subparser(func: typing.Callable[[model.Api, argparse.Namespace], None], *, name: str='') -> argparse.ArgumentParser:
		if not name:
			name = func.__name__
		parser = subparsers.add_parser(name, help=func.__doc__, description=func.__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
		parser.set_defaults(func=func)
		return parser

	WARNING_COMMA = "(note that commas must be escaped in sway config)"

	root_parser.set_defaults(func=default, time_in_s=DEFAULT_TIME_IN_S)
	root_parser.add_argument('--version', action='store_true', help=print_version.__doc__)
	root_parser.add_argument('-n', '--dry-run', action='store_true', help='do not run the command to change the monitors (implies --verbose)')
	root_parser.add_argument('-v', '--verbose', action='store_true', help='print the command to change the monitors')
	root_parser.add_argument('--log', metavar='FILE', help='redirect stdout and stderr to %(metavar)s (implies --verbose)')
	root_parser.add_argument('--test-input', help='a file which contains the output of the backend')
	root_parser.add_argument('--backend', default=DEFAULT_BACKEND, choices=BACKENDS.keys(), help='the command used to configure the monitors')
	root_parser.add_argument('--use-short-names-on-sway', action='store_true', help='[swaymsg only] use names like eDP-1 and DP-4. These names can change when disconnecting and reconnecting a monitor which can lead to a situation where all monitors are disabled even after reconnecting a previously enabled monitor. Therefore the default behavior is to use longer names based on make, model and serial, see man sway-output.')
	root_parser.add_argument('--wait', type=float, help='time to wait before starting, this may help if you are experiencing trouble when binding this command to a key on sway')

	parser = add_subparser(cycle)
	parser.add_argument('-t', '--time-in-s', type=float, default=DEFAULT_TIME_IN_S, help='time in seconds to wait between switching the monitor and switching back to the previously focused workspace (sway only, i3 does not switch workspaces when switching monitors)')
	parser.add_argument('connections', type=connection, nargs='?', default=CONNECTION_ANY, help=f'a comma separated list of connections which should be considered for activation: {possible_connections} {WARNING_COMMA}')

	parser = add_subparser(toggle)
	parser.add_argument('-t', '--time-in-s', type=float, default=DEFAULT_TIME_IN_S, help='time in seconds to wait between switching the monitor and switching back to the previously focused workspace (sway only, i3 does not switch workspaces when switching monitors)')
	parser.add_argument('connections1', type=connection, nargs='?', default=CONNECTION_EXTERNAL, help=f'a comma separated list of {possible_connections} {WARNING_COMMA}')
	parser.add_argument('connections2', type=connection, nargs='?', default=model.CONNECTION_INTERNAL, help=f'a comma separated list of {possible_connections} {WARNING_COMMA}')

	parser = add_subparser(extend)
	parser.add_argument('direction', type=model.Direction, help='one of %s' % ', '.join(d.value for d in model.Direction))
	parser.add_argument('connections', type=connection, nargs='?', default=connection('external,any'), help=f'a comma separated list of connections which should be considered for activation: {possible_connections} {WARNING_COMMA}')

	parser = add_subparser(mirror)
	parser.add_argument('-o', '--original', type=connection, default=CONNECTION_EXTERNAL, help='the content on this screen is always displayed at 100%%, if the other screen has a different resolution it will be scaled there')
	parser.add_argument('connections', type=connection, nargs='?', default=CONNECTION_ANY, help=f'a comma separated list of connections which should be considered for activation: {possible_connections} {WARNING_COMMA}')

	parser = add_subparser(scale)
	parser.add_argument('scale', type=model.Scale.parse, help=f'<scale_x>[x<scale_y>], if scale_y is omitted it is set to scale_x. If you pass a 2 everything on the monitor will be half as big so that twice as much will fit on it. Pass 1 to return to normal. Sway does not support different scale values for x and y direction.')
	parser.add_argument('connections', type=connection, nargs='?', default=CONNECTION_ANY, help=f'a comma separated list of connections which should be scaled: {possible_connections} {WARNING_COMMA}')

	parser = add_subparser(rotate)
	parser.add_argument('rotation', type=model.Rotation, help='one of %s' % ', '.join(r.value for r in model.Rotation))
	parser.add_argument('connections', type=connection, nargs='?', default=CONNECTION_ANY, help=f'a comma separated list of connections which should be rotated: {possible_connections} {WARNING_COMMA}')

	parser = add_subparser(reset)
	parser.add_argument('-t', '--time-in-s', type=float, default=DEFAULT_TIME_IN_S, help='time in seconds to wait between switching the monitor and switching back to the previously focused workspace (sway only, i3 does not switch workspaces when switching monitors)')
	parser.add_argument('connections', type=connection, nargs='?', default=CONNECTION_ANY, help=f'a comma separated list of connections out of which the first is kept on: {possible_connections} {WARNING_COMMA}')

	parser = add_subparser(ls, name='list')
	parser.add_argument('connections', type=connection, nargs='?', default=CONNECTION_ANY, help=f'a comma separated list of connections which should be printed: {possible_connections} {WARNING_COMMA}')
	parser.add_argument('-f', '--format', default='{m.name:8} {"("m.connection")":12}  {"["m.on_str"]":3}  {m.default_resolution!s:10} {m.long_name}', help=f'the pattern of the lines to be printed')

	parser = add_subparser(config)
	parser.add_argument('-l', '--list-searched-paths', action='store_true', help='list all config file names which are checked for existence')
	parser.add_argument('-p', '--print-file-name', action='store_true', help='print the path of the config file or nothing if no config file is found')
	parser.add_argument('-e', '--edit', action='store_true', help='open the config file in EDITOR')

	#parser = subparsers.add_parser('load', help='load a configuration')
	#parser.set_defaults(func=load)

	return root_parser

def get_api(args: argparse.Namespace) -> model.Api:
	name: str = args.backend
	module = BACKENDS[name]
	if module:
		api_module = importlib.import_module(module, APP_NAME)
	else:
		for module in BACKENDS.values():
			if not module:
				continue
			api_module = importlib.import_module(module, APP_NAME)
			if api_module.is_available():
				break
		else:
			print('Failed to find an available backend. The following are supported but none of them seems to be installed:', file=sys.stderr)
			for backend in BACKENDS:
				print(f'- {backend}', file=sys.stderr)
			sys.exit(1)
	
	api: model.Api = api_module.Api(**vars(args))
	return api

def main(command_line_args: typing.Optional[typing.Sequence[str]]=None) -> None:
	parser = create_parser()
	args = parser.parse_args(command_line_args)
	if args.wait:
		time.sleep(args.wait)
	if args.dry_run:
		args.verbose = True

	original_stdout = sys.stdout
	original_stderr = sys.stderr
	logfile: 'typing.IO[str]|None' = None
	try:
		if args.log:
			args.verbose = True
			logfile = open(args.log, 'wt')
			sys.stdout = logfile
			sys.stderr = logfile
			import datetime
			print(datetime.datetime.now())
			print(shlex.join(sys.argv))
		func: typing.Callable[[model.Api, argparse.Namespace], None] = args.func
		if func is not config:
			parse_config()
		if args.version:
			print_version(args)
		api = get_api(args)
		try:
			func(api, args)
		except model.NotSupportedException as e:
			print(e, file=sys.stderr)
			sys.exit(1)
	finally:
		sys.stderr = original_stderr
		sys.stdout = original_stdout
		if logfile:
			logfile.close()


if __name__ == '__main__':
	main()
